/*
	Warsztaty OneWire
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include "fxdebug.h"


#define STATE_FINISHED 0
#define STATE_READ_RESET 1
#define STATE_READ_PRESENCE_WAIT 2
#define STATE_READ_COMMAND_BS 3
#define STATE_READ_COMMAND_BE 4
#define STATE_READ_COMMAND_WAIT 5
#define STATE_READ_BYTE 6
#define STATE_READ_BYTE_2 7
#define STATE_READ_BYTE_3 8

#define COMMAND_SKIP_ROM_CONVERT_NEXT (1 << 7)
#define COMMAND_SKIP_ROM_READ_NEXT (1 << 6)
#define COMMAND_CONVERT (1 << 5)
#define COMMAND_READ (1 << 4)
#define WAITING_UNTIL_READY (1 << 7)

#define ONEWIRE_PIN PC0
#define DEBUG_PIN PC1

//set this to 1 if you have DS18B20 instead of DS1820
#define DS18B20 0

volatile unsigned char bitsRemaining; //also used as COMMAND_ state
volatile unsigned char state = STATE_FINISHED;

volatile unsigned char commandBuffer = 0;

volatile unsigned temp1;
volatile unsigned temp2;

void scheduleTrigger(const char ticks) {
	//setup timer0
	TCCR0 = (1 << CS00) | (1 << CS01) | (1 << WGM01);
	TCNT0 = 0;
	OCR0 = ticks;
	TIMSK = 1 << OCIE0;
}

void scheduleFastTrigger(const char ticks) {
	//setup timer0
	TCCR0 = (1 << CS01) | (1 << WGM01);
	TCNT0 = 0;
	OCR0 = ticks;
	TIMSK = 1 << OCIE0;
}

ISR(TIMER0_COMP_vect) {
	//disable this timer for now
	TCCR0 = 0;

	//do stuff depending on state
	if (state == STATE_READ_RESET) {
		//release one-wire bus
		DDRC &= ~(1 << ONEWIRE_PIN);
		state = STATE_READ_PRESENCE_WAIT;
		scheduleTrigger(9);
		return;
	}


	if (state == STATE_READ_PRESENCE_WAIT) {
		//check if low
		PORTC |= (1 << DEBUG_PIN);
		PORTC &=~ (1 << DEBUG_PIN);
		if ( ! (PINC & (1 << ONEWIRE_PIN)) ) {
			//device present!
			state = STATE_READ_COMMAND_BS;

			commandBuffer = 0xCC; //SKIP ROM command
			if (bitsRemaining == 0)
				bitsRemaining = 8 | COMMAND_SKIP_ROM_CONVERT_NEXT;
			else bitsRemaining = 8 | COMMAND_SKIP_ROM_READ_NEXT;
			//wait the rest of time
			scheduleTrigger(60);
		}
		return;
	}

	if (state == STATE_READ_COMMAND_BS) {
		state = STATE_READ_COMMAND_WAIT;
		DDRC |= 1 << ONEWIRE_PIN;

		if (commandBuffer & 1) {
			//write 1 - hold bus low shorter
			scheduleTrigger(1);
		}
		else {
			//write 0 byte - hold bus low longer
			scheduleTrigger(15);
		}
		return;

	}

	if (state == STATE_READ_COMMAND_WAIT) {
		DDRC &= ~(1 << ONEWIRE_PIN); //release bus
		state = STATE_READ_COMMAND_BE;
		scheduleTrigger(12); //finalize
		return;
	}

	if (state == STATE_READ_COMMAND_BE) {

		commandBuffer >>= 1;
		state = STATE_READ_COMMAND_BS;

		if (!((--bitsRemaining) & 0x0F)) { //if there aren't any remaining bytes
			if (bitsRemaining & COMMAND_SKIP_ROM_CONVERT_NEXT) {
				commandBuffer = 0x44; //CONVERT command
				bitsRemaining = 8 | COMMAND_CONVERT;
			}
			else if (bitsRemaining & COMMAND_SKIP_ROM_READ_NEXT) {
				commandBuffer = 0xBE; //READ SCRATCHPAD command
				bitsRemaining = 8 | COMMAND_READ;
			}
			else if (bitsRemaining & COMMAND_CONVERT) {
				state = STATE_READ_BYTE;
				bitsRemaining = WAITING_UNTIL_READY; //special state, not bits count
			}
			else if (bitsRemaining & COMMAND_READ) {
				state = STATE_READ_BYTE;
				bitsRemaining = 16; //need to read 16 bits (2 bytes) from scratchpad
			}
		}
		//delay for rising edge
		scheduleTrigger(2); //16us
		return;
	}

	if (state == STATE_READ_BYTE) {
		//generate slot
		if (bitsRemaining == 0) {
			//transmit
            state = STATE_FINISHED;
            scheduleTrigger(1);
			return;
		}

		DDRC |= 1 << ONEWIRE_PIN;
		state = STATE_READ_BYTE_2;
		scheduleFastTrigger(3);
		return;
	}

	if (state == STATE_READ_BYTE_2) {
		DDRC &= ~(1 << ONEWIRE_PIN); //release bus
		state = STATE_READ_BYTE_3;
		scheduleFastTrigger(10);
		return;
	}

	if (state == STATE_READ_BYTE_3) {

		if (bitsRemaining == WAITING_UNTIL_READY) {
			PORTC |= (1 << DEBUG_PIN);
			PORTC &= ~(1 << DEBUG_PIN);

			if (PINC & (1 << ONEWIRE_PIN)) {
				//ready!
				//reset
				DDRC |= 1 << ONEWIRE_PIN;

				state = STATE_READ_RESET;
				bitsRemaining = 1; //so it'll do READ instead of CONVERT next time
				scheduleTrigger(200);
				return;
			}
			else {
				//not ready
				state = STATE_READ_BYTE;
				scheduleTrigger(30);
				return;
			}
		}
		else {
			bitsRemaining--;

			if (bitsRemaining > 7) {
				temp1 >>= 1;
				PORTC |= (1 << DEBUG_PIN);
				PORTC &=~ (1 << DEBUG_PIN);

				if (PINC & (1 << ONEWIRE_PIN)) {
					temp1 |= (1 << 7);
				}
			}
			else {
				temp2 >>= 1;

				PORTC |= (1 << DEBUG_PIN);
				PORTC &=~ (1 << DEBUG_PIN);

				if (PINC & (1 << ONEWIRE_PIN)) {
					temp2 |= (1 << 7);
				}
			}
			state = STATE_READ_BYTE;
			scheduleTrigger(30);
			return;
		}
		return;
	}


}


int main(void)
{
	stdout=&fxStream;
	sei();
	printf("\nStart!");

	PORTC = 0;
	
	//debug pin is used only to check program state on logic analyzer
	//you can safely comment all lines with DEBUG_PIN
	DDRC = (1 << DEBUG_PIN);

	while(1) {
		if (state == STATE_FINISHED) {
			unsigned char integ = 0;
			unsigned char sign = 0;
			char fraction[4] = {'0', '0', '0', '0'};
			
			//if we're serious (and want compact code), we should avoid floats by all means
			//so we need to do some work to correctly decode readings
			uint16_t tempFull = (temp2 << 8) | temp1;

			if (tempFull & (1 << 15)) {
				sign = 1;
				tempFull -= 1;
				tempFull ^= 0xFFFF;
			}

			if (DS18B20) {
				integ = (tempFull >> 4);

				if (tempFull & (1 << 0)) {
					fraction[1] += 6;
					fraction[2] += 2;
					fraction[3] += 5;
				}
				if (tempFull & (1 << 1)) {
					fraction[0] += 1;
					fraction[1] += 2;
					fraction[2] += 5;
					//fraction[3] += 0;
				}
				if (tempFull & (1 << 2)) {
					fraction[0] += 2;
					fraction[1] += 5;
					//fraction[2] += 0;
					//fraction[3] += 0;
				}

				if (tempFull & (1 << 3)) {
					fraction[0] += 5;
					//fraction[1] += 0;
					//fraction[2] += 0;
					//fraction[3] += 0;
				}

				if (fraction[1] > '9') {
					fraction[1] -= 10;
					fraction[0] += 1;
				}


			}
			else {
				//ds1820
				integ = (tempFull >> 1);
				if (tempFull & 1) {
					fraction[0] += 5;
				}
			}

			printf("\n\n0x%X 0x%X\n", temp1, temp2);
			if (sign) printf("-");
			printf("%i.%c",integ, fraction[0]);

			if (fraction[1] != '0')
				printf("%c", fraction[1]);

			if (fraction[2] != '0')
				printf("%c", fraction[2]);

			if (fraction[3] != '0')
				printf("%c", fraction[3]);

			printf("'C");


			DDRC |= 1 << ONEWIRE_PIN;
			state = STATE_READ_RESET;
			bitsRemaining = 0;
			scheduleTrigger(200);

		}

    }

    return 0;
}
